#!/usr/bin/env python
# coding: utf-8

"""
Converts a native Python dictionary into an XML string. Supports int, float, str, unicode, list, dict and arbitrary nesting.
"""

from __future__ import unicode_literals

__version__ = '1.5.6'
version = __version__

from random import randint
import re
import collections
import logging
from xml.dom.minidom import parseString

# python 3 doesn't have a unicode type
try:
    unicode
except:
    unicode = str

# python 3 doesn't have a long type
try:
    long
except:
    long = int


def set_debug(debug=True, filename='dicttoxml.log'):
    if debug:
        import datetime
        print('Debug mode is on. Events are logged at: %s' % filename)
        logging.basicConfig(filename=filename, level=logging.INFO)
        logging.info('\nLogging session starts: %s' % (str(datetime.datetime.today())))
    else:
        logging.basicConfig(level=logging.WARNING)
        print('Debug mode is off.')

ids = [] # initialize list of unique ids


def make_id(element, start=100000, end=999999):
    """Returns a random integer"""
    return '%s_%s' % (element, randint(start, end))


def get_unique_id(element):
    """Returns a unique id for a given element"""
    this_id = make_id(element)
    dup = True
    while dup is True:
        if this_id not in ids:
            dup = False
            ids.append(this_id)
        else:
            this_id = make_id(element)
    return ids[-1]


def get_xml_type(val):
    """Returns the data type for the xml type attribute"""
    if type(val).__name__ in ('str', 'unicode'):
        return 'str'
    if type(val).__name__ in ('int', 'long'):
        return 'int'
    if type(val).__name__ == 'NoneType':
        return 'null'
    if isinstance(val, dict):
        return 'dict'
    if type(val).__name__ in ('list', 'set', 'tuple') or isinstance(val, collections.Iterable):
        return 'list'
    return type(val).__name__


def xml_escape(s):
    if type(s) in (str, unicode):
        s = s.replace('&', '&amp;')
        s = s.replace('"', '&quot;')
        s = s.replace('\'', '&apos;')
        s = s.replace('<', '&lt;')
        s = s.replace('>', '&gt;')
    return s


def make_attrstring(attr):
    """Returns an attribute string in the form key="val" """
    attrstring = ' '.join(['%s="%s"' % (k, v) for k, v in attr.items()])
    return '%s%s' % (' ' if attrstring != '' else '', attrstring)


def key_is_valid_xml(keys):
    if not type(keys) is tuple:
        keys = (keys, keys)

    """Checks that a keys is a valid XML name"""
    logging.info('Inside keys_is_valid_xml(). Testing "<%s></%s>"' % (keys[0], keys[1]))
    test_xml = '<?xml version="1.0" encoding="UTF-8" ?><%s>foo</%s>' % (keys[0], keys[1])
    try: 
        parseString(test_xml)
        return True
    except Exception: #minidom does not implement exceptions well
        return False


def make_valid_xml_name(key, attr):
    if type(key) is tuple:
        key = key[0]

    """Tests an XML name and fixes it if invalid"""
    logging.info('Inside make_valid_xml_name(). Testing key "%s" with attr "%s"' % (key, str(attr)))
    # pass through if key is already valid
    if key_is_valid_xml(key):
        return (key, key), attr
    # prepend a lowercase n if the key is numeric
    if key.isdigit():
        key = 'n%s' % key
        return (key, key), attr
    # replace spaces with underscores if that fixes the problem
    if key_is_valid_xml(key.replace(' ', '_')):
        key = key.replace(' ', '_')
        return (key, key), attr

    # Check if attributes were added to key, if so strip from closed key
    closed_key = re.sub(r'(\S+)\s+.*', r'\1', key)
    if key_is_valid_xml((key, closed_key)):
        return (key, closed_key), attr

    # key is still invalid - move it into a name attribute
    attr['name'] = key
    key = 'key'
    return (key, key), attr


def convert(obj, ids, attr_type, parent='root'):
    """Routes the elements of an object to the right function to convert them based on their data type"""
    logging.info('Inside convert(). obj type is: "%s", obj="%s"' % (type(obj).__name__, obj))
    if type(obj) in (int, float, long, str, unicode):
        return convert_kv('item', obj, attr_type)
    if hasattr(obj, 'isoformat'):
        return convert_kv('item', obj.isoformat(), attr_type)
    if type(obj) == bool:
        return convert_bool('item', obj, attr_type)
    if obj is None:
        return convert_none('item', '', attr_type)
    if isinstance(obj, dict):
        return convert_dict(obj, ids, parent, attr_type)
    if type(obj) in (list, set, tuple) or isinstance(obj, collections.Iterable):
        return convert_list(obj, ids, parent, attr_type)
    raise TypeError('Unsupported data type: %s (%s)' % (obj, type(obj).__name__))


def convert_dict(obj, ids, parent, attr_type):
    """Converts a dict into an XML string."""
    logging.info('Inside convert_dict(): obj type is: "%s", obj="%s"' % (type(obj).__name__, obj))
    output = []
    addline = output.append
    for key, val in obj.items():
        logging.info('Looping inside convert_dict(): key="%s", val="%s", type(val)="%s"' % (key, val, type(val).__name__))

        this_id = get_unique_id(parent)
        attr = {} if ids == False else {'id': '%s' % this_id }
        
        keys, attr = make_valid_xml_name(key, attr)
       
        if type(val) in (int, float, long, str, unicode):
            addline(convert_kv(keys, val, attr_type, attr))
            
        elif hasattr(val, 'isoformat'):  # datetime
            addline(convert_kv(keys, val.isoformat(), attr_type, attr))
            
        elif type(val) == bool:
            addline(convert_bool(keys, val, attr_type, attr))
            
        elif isinstance(val, dict):
            if attr_type:
                attr['type'] = get_xml_type(val)
            addline('<%s%s>%s</%s>' % (
                keys[0], make_attrstring(attr), convert_dict(val, ids, keys[0], attr_type), keys[1])
            )
        elif type(val) in (list, set, tuple) or isinstance(val, collections.Iterable):
            if attr_type:
                attr['type'] = get_xml_type(val)
            addline('<%s%s>%s</%s>' % (
                keys[0], make_attrstring(attr), convert_list(val, ids, keys[0], attr_type), keys[1])
            )
        elif val is None:
            addline(convert_none(keys, val, attr_type, attr))
        else:
            raise TypeError('Unsupported data type: %s (%s)' % (obj, type(obj).__name__))
    return ''.join(output)


def convert_list(items, ids, parent, attr_type):
    """Converts a list into an XML string."""
    logging.info('Inside convert_list()')
    output = []
    addline = output.append
    this_id = get_unique_id(parent)
    for i, item in enumerate(items):
        logging.info('Looping inside convert_list(): item="%s", type="%s"' % (item, type(item).__name__))
        attr = {} if ids == False else {
            'id': '%s_%s' % (this_id, i+1) 
        }
        if type(item) in (int, float, long, str, unicode):
            addline(convert_kv('item', item, attr_type, attr))
        elif hasattr(item, 'isoformat'):  # datetime
            addline(convert_kv('item', item.isoformat(), attr_type, attr))
        elif type(item) == bool:
            addline(convert_bool('item', item, attr_type, attr))
        elif isinstance(item, dict):
            if not attr_type:
                addline('<item>%s</item>' % (convert_dict(item, ids, parent, attr_type)))
            else:
                addline('<item type="dict">%s</item>' % (convert_dict(item, ids, parent, attr_type)))
        elif type(item) in (list, set, tuple) or isinstance(item, collections.Iterable):
            if not attr_type:
                addline('<item %s>%s</item>' % (make_attrstring(attr), convert_list(item, ids, 'item', attr_type)))
            else:
                addline('<item type="list"%s>%s</item>' % (make_attrstring(attr), convert_list(item, ids, 'item', attr_type)))
        elif item is None:
            addline(convert_none('item', None, attr_type, attr))
        else:
            raise TypeError('Unsupported data type: %s (%s)' % (item, type(item).__name__))
    return ''.join(output)


def convert_kv(keys, val, attr_type, attr={}):
    """Converts an int, float or string into an XML element"""
    logging.info('Inside convert_kv(): keys="%s", val="%s", type(val) is: "%s"' % (str(keys), val, type(val).__name__))

    # Don't normalize twice
    if not keys is tuple:
        keys, attr = make_valid_xml_name(keys, attr)
    
    if attr_type:
        attr['type'] = get_xml_type(val)
    attrstring = make_attrstring(attr)

    return '<%s%s>%s</%s>' % (
        keys[0], attrstring, xml_escape(val), keys[1]
    )


def convert_bool(keys, val, attr_type, attr={}):
    """Converts a boolean into an XML element"""
    logging.info('Inside convert_bool(): keys="%s", val="%s", type(val) is: "%s"' % (str(keys), val, type(val).__name__))

    # Don't normalize twice
    if not keys is tuple:
        keys, attr = make_valid_xml_name(keys, attr)
    
    if attr_type:
        attr['type'] = get_xml_type(val)
    attrstring = make_attrstring(attr)
    return '<%s%s>%s</%s>' % (keys[0], attrstring, unicode(val).lower(), keys[1])


def convert_none(keys, val, attr_type, attr={}):
    """Converts a null value into an XML element"""
    logging.info('Inside convert_none(): keys="%s"' % str(keys))

    # Don't normalize twice
    if not keys is tuple:
        keys, attr = make_valid_xml_name(keys, attr)

    if attr_type:
        attr['type'] = get_xml_type(val)
    attrstring = make_attrstring(attr)
    return '<%s%s></%s>' % (keys[0], attrstring, keys[1])


def dicttoxml(obj, root=True, custom_root='root', ids=False, attr_type=True):
    """Converts a python object into XML
    attr_type is used to specify if data type for each element should be included in the resulting xml.
    By default, it is set to True.
    """
    logging.info('Inside dicttoxml(): type(obj) is: "%s", obj="%s"' % (type(obj).__name__, obj))
    output = []
    addline = output.append
    if root is True:
        addline('<?xml version="1.0" encoding="UTF-8" ?>')
        root_keys, attr = make_valid_xml_name(custom_root, {})
        addline('<%s>%s</%s>' % (root_keys[0], convert(obj, ids, attr_type, parent=custom_root), root_keys[1]))
    else:
        addline(convert(obj, ids, attr_type, parent=''))
    return ''.join(output).encode('utf-8')


HTTP_STATUS_OK = 200
HTTP_STATUS_MOVED = 302
HTTP_STATUS_BAD_REQUEST = 400
HTTP_STATUS_FORBIDDEN = 403
HTTP_STATUS_NOT_FOUND = 404

HTTP_DEFAULT_PORT = 80
HTTPS_DEFAULT_PORT = 443

URL_EXPIRATION_IN_SECONDS = 86400
MAX_AUTH_ATTEMPTS = 2

BUSINESS_OWNER_PROFILE_ID = 'BusinessProfile'
SBCS_ID = 'SmallBusinessCreditShare'
BUSINESS_PREMIER_PROFILE_ID = 'PremierProfile'

TERMS_RESPONSE_CODE = '031'



from pyexperian import services


def test_get_hostname_from_url():
    assert services.Ecals.get_hostname_from_url('https://www.google.com/blog') == 'www.google.com'
    assert services.Ecals.get_hostname_from_url('http://candy-crush.blah-3.org') == 'candy-crush.blah-3.org'
    assert services.Ecals.get_hostname_from_url('candy-crush.blah') is None


def test_net_connect_url():
    assert not services.Ecals.is_valid_net_connect_url('https://experian.fake.com')
    assert not services.Ecals.is_valid_net_connect_url('https://experian.org')
    assert services.Ecals.is_valid_net_connect_url('https://dm1.experian.com/netconnect2_0Demo/servlets/NetConnectServlet')


def test_fetch_net_connect_url():
    ecals = services.Ecals("http://www.experian.com/lookupServlet1?lookupServiceName=AccessPoint&lookupServiceVersion=1.0&serviceName=NetConnectDemo&serviceVersion=2.0&responseType=text/plain")
    assert ecals.get_net_connect_url() is not None



from pyexperian import services, exceptions
from nose.tools import raises

try:
    from pyexperian.test import secrets

    ecals = services.Ecals(secrets.ECALS_URL)
    bpp = services.BusinessPremierProfile(secrets.config, ecals)
    sbcs = services.SBCS(secrets.config, ecals)
    bop = services.BusinessOwnerProfile(secrets.config, ecals)
except ImportError as e:
    ecals = bpp = sbcs = bop = None


@raises(exceptions.IncompleteBusinessException)
def test_empty_business():
    global bpp

    if not bpp:
        return

    bpp.query(business={})


@raises(exceptions.IncompleteBusinessException)
def test_no_business_address():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'Experian Information Solutions'
    }

    bpp.query(business=business)


@raises(exceptions.IncompleteBusinessException)
def test_incomplete_business_address():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'state': 'CA',
            'zip': '92626'
        }
    }

    bpp.query(business=business)


def test_complete_business_address():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'Experian Information Solutiosn',
        'address': {
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    resp = bpp.query(business=business)


    assert resp is not None



@raises(exceptions.IncompleteOwnerException)
def test_empty_owner():
    global bop

    if not bop:
        return

    owner = {

    }

    bop.query(owner=owner)


@raises(exceptions.IncompleteOwnerException)
def test_incomplete_owner_name():
    global bop

    if not bop:
        return

    owner = {
        'last_name': 'Benson',
        'address': {
            'street': '7600 Trumbower Trl',
            'city': 'Millington',
            'state': 'MI',
            'zip': '48746'
        }
    }

    bop.query(owner=owner)



@raises(exceptions.IncompleteOwnerException)
def test_incomplete_address():
    global bop

    if not bop:
        return

    owner = {
        'first_name': 'Derrick',
        'last_name': 'Benson',
        'address': {
            'city': 'Millington',
            'state': 'MI',
            'zip': '48746'
        }
    }

    bop.query(owner=owner)


def test_complete_owner():
    global bop

    if not bop:
        return

    business = {
        'name': 'Experian Information Solutiosn',
        'address': {
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    owner = {
        'first_name': 'Derrick',
        'last_name': 'Benson',
        'address': {
            'street': '7600 Trumbower Trl',
            'city': 'Millington',
            'state': 'MI',
            'zip': '48746'
        }
    }

    resp = bop.query(business=business, owner=owner)


    assert resp is not None


from pyexperian import services, parsers, exceptions
from nose.tools import raises

try:
    from pyexperian.test import secrets

    ecals = services.Ecals(secrets.ECALS_URL)
    bpp = services.BusinessPremierProfile(secrets.config, ecals)
    sbcs = services.SBCS(secrets.config, ecals)
    bop = services.BusinessOwnerProfile(secrets.config, ecals)
    raw = services.Raw(secrets.config, ecals)
except ImportError as e:
    ecals = bpp = sbcs = bop = raw = None


def test_no_hit_premier_profile():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'NORECORDCO',
        'address': {
            'street': '123 Main Street',
            'city': 'Buena Park',
            'state': 'CA',
            'zip': '90620'
        }
    }

    bpp = services.BusinessPremierProfile(secrets.config, ecals)

    resp_blob = bpp.query(business=business)

    assert resp_blob is not None

    result = parsers.BusinessPremierProfile(resp_blob)

    assert not result.business_found()


def test_direct_hit_premier_profile():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'street': '475 Anton Blvd',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    resp_blob = bpp.query(business=business)

    result = parsers.BusinessPremierProfile(resp_blob)

    assert not result.has_list()
    assert result.business_found()
    business_info = result.get_business()
    assert 'experian_bin' in business_info
    assert 'phone' in business_info
    assert 'address' in business_info

def test_list_of_similars_premier_profile():
    global bpp

    if not bpp:
        return

    business = {
        'name': 'Experian',
        'address': {
            'street': '123 Main Street',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    resp_blob = bpp.query(business=business)

    result = parsers.BusinessPremierProfile(resp_blob)

    assert result.has_list()

    similars = result.get_list()

    assert len(similars) > 1


    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'street': '123 Main Street',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    resp_blob = bpp.query(business=business)

    result = parsers.BusinessPremierProfile(resp_blob)

    assert result.has_list()
    similars = result.get_list()
    similar = similars[0]

    assert len(similars) == 1
    assert 'experian_bin' in similar
    assert 'name' in similar
    assert 'address' in similar

    business = {
        'experian_bin': similar['experian_bin']
    }

    resp_blob = bpp.query(business=business)

    result = parsers.BusinessPremierProfile(resp_blob)

    assert result.business_found()
    business = result.get_business()

    assert business['name'] == similar['name']


def test_standalone_business_owner_profile():
    global bop

    if not bop:
        return

    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'street': '475 Anton Blvd',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    owner = {
        'first_name': 'Derrick',
        'last_name': 'Benson',
        'ssn': '887487109',
        'address': {
            'street': '7600 Trumbower Trl',
            'city': 'Millington',
            'state': 'MI',
            'zip': '48746'
        }
    }

    resp_blob = bop.query(business=business, owner=owner)

    result = parsers.BusinessOwnerProfile(resp_blob)

    assert result.owner_found()

def test_nohit_standalone_business_owner_profile():
    global bop

    if not bop:
        return

    business = {
        'name': 'Fake Biz',
        'phone': '999-555-1234',
        'address': {
            'street': 'Fake Addy',
            'city': 'Costa Fake',
            'state': 'CA',
            'zip': '78734'
        }
    }

    owner = {
        'first_name': 'Santa',
        'last_name': 'Jackson',
        'ssn': '999999999',
        'address': {
            'street': 'Fake Addy2',
            'city': 'Texas',
            'state': 'CA',
            'zip': '99911'
        }
    }

    resp_blob = bop.query(business=business, owner=owner)

    result = parsers.BusinessOwnerProfile(resp_blob)

    assert not result.business_found()
    assert not result.owner_found()

def test_no_hit_sbcs():
    global sbcs

    if not sbcs:
        return

    business = {
        'name': 'NORECORDCO',
        'address': {
            'street': '123 Main Street',
            'city': 'Buena Park',
            'state': 'CA',
            'zip': '90620'
        }
    }

    resp_blob = sbcs.query(business=business)

    result = parsers.SBCS(resp_blob)

    assert not result.business_found()


def test_direct_hit_sbcs():
    global sbcs

    if not sbcs:
        return

    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'street': '475 Anton Blvd',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }



    resp_blob = sbcs.query(business=business)

    result = parsers.SBCS(resp_blob)

    assert result.business_found()


def test_list_of_similars_sbcs():
    global sbcs

    if not sbcs:
        return

    business = {
        'name': 'Experian Information Solutions',
        'address': {
            'street': '123 Main Street',
            'city': 'Costa Mesa',
            'state': 'CA',
            'zip': '92626'
        }
    }

    resp_blob = sbcs.query(business=business)

    result = parsers.SBCS(resp_blob)

    assert result.has_list()


def test_raw_query():
    global raw

    if not raw:
        return

    xml = """
    <?xml version="1.0" encoding="UTF-8"?>
    <NetConnectRequest>
        <EAI>%(eai)s</EAI>
        <DBHost>%(db_host)s</DBHost>
        <Request xmlns="http://www.experian.com/WebDelivery" version="1.0">
            <Products>
                <PremierProfile>
                    <Subscriber>
                        <OpInitials>OP</OpInitials>
                        <SubCode>%(sub_code)s</SubCode>
                    </Subscriber>
                    <BusinessApplicant>
                        <BusinessName>EXPERIAN INFORMATION SOLUTIONS</BusinessName>
                        <CurrentAddress>
                            <Street>475 ANTON BLVD</Street>
                            <City>COSTA MESA</City>
                            <State>CA</State>
                            <Zip>92626</Zip>
                        </CurrentAddress>
                    </BusinessApplicant>
                    <OutputType>
                        <XML>
                            <Verbose>Y</Verbose>
                        </XML>
                    </OutputType>
                    <Vendor>
                        <VendorNumber>$(vendor_number)s</VendorNumber>
                    </Vendor>
                </PremierProfile>
            </Products>
        </Request>
    </NetConnectRequest>
    """ % {
            'eai': secrets.EAI,
            'db_host': secrets.DB_HOST,
            'sub_code': secrets.SUB_CODE,
            'vendor_number': secrets.VENDOR_NUMBER
        }

    result = raw.query(xml)

    assert result.index('NetConnectResponse') >= 0

@raises(exceptions.BadRequestException)
def test_bad_request_query():
    global raw

    if not raw:
        return

    xml = """
    <?xml version="1.0" encoding="UTF-8"?>
    <NetConnectRequest>
        <EAI>%(eai)s</EAI>
        <DBHost>%(db_host)s</DBHost>
        <Request xmlns="http://www.experian.com/WebDelivery" version="1.0">
            <Products>
                <PremierProfile>
                    <Subscriber>
                        <OpInitials>OP</OpInitials>
                        <SubCode>%(sub_code)s</SubCode>
                    </Subscriber>
                    <BusinessApplicant>
                        <BusinessName>EXPERIAN INFORMATION SOLUTIONS</BusinessName>
                        <CurrentAddress>
                            <Zip>92626</Zip>
                        </CurrentAddress>
                    </BusinessApplicant>
                    <OutputType>
                        <XML>
                            <Verbose>Y</Verbose>
                        </XML>
                    </OutputType>
                    <Vendor>
                        <VendorNumber>$(vendor_number)s</VendorNumber>
                    </Vendor>
                </PremierProfile>
            </Products>
        </Request>
    </NetConnectRequest>
    """ % {
            'eai': secrets.EAI,
            'db_host': secrets.DB_HOST,
            'sub_code': secrets.SUB_CODE,
            'vendor_number': secrets.VENDOR_NUMBER
        }

    result = raw.query(xml)

    parsed_result = parsers.BusinessPremierProfile(result)
